;;; File: env_M+1auction.lisp -*- Mode: Lisp; Syntax: Common-Lisp; -*-
;;; Contine functiile:

;;;   (clear_auction good env deadline)
;;;   (contract_generation sell buy M+1_price good env deadline)


;;;;*******************************
;;;; Clear auction
;;;;*******************************

;;; 1) se  lucreaza pe copii ale listelor cu ofertele de vanzare si cumparare
;;;    din runda curenta
;;; 2) toate ofertele se pun in lista all care are forma ( (a1. 2) (a3. 4)...)
;;; 3) se ordoneaza aceasta lista de asocatii (=> all)
;;; 3) daca nu exista oferte de vanzare pretul este 0
;;;    daca exista oferte de vanzare se extrage din lista ordonata all al M+1
;;;    prt (=> M+1_price)
;;; 4) se sorteaza crescator ofertele de vanzare si cumparare (=> sell, buy)         
;;; 5) se apeleaza functia (contract_generation sell buy M+1_price good env
;;;    deadline)
 
(defun clear_auction (good env deadline)
"clear algorithm for M+1_st price auction"
  (setq buy (copy-list (good-buy_bids good))
  	  sell (copy-list (good-sell_bids good))
  	  s_copy (copy-list sell)
  	  b_copy (copy-list buy)
  	  all (append s_copy b_copy)
  	  M (length sell))
  (format t "~% All bids for good ~A are ~S " (good-name good) all )
  (format t "~% ~A oferte de vanzare pentru bunul ~A " M (good-name good))
  (setf all (sort_assoc all))
  ;;daca exista oferte de vanzare se stabilste pretul, altfel prt=0
  (cond ((> M 0) (setq M+1_price (cdr (nth (- M 1) all))))
        (T (format t "~% Nu exista oferte de vanzare pentru bunul ~A" 
           (good-name good)) 
           (setf M+1_price 0)))
  (format t "~% Pretul este ~S" M+1_price)
  (setf buy (sort_assoc buy))
  (format t " ~% ofertele de cumparare sunt: ~S" buy)
  (setf sell (sort_assoc sell))
  (format t " ~% ofertele de vanzare sunt: ~S" sell)

  (contract_generation sell buy M+1_price good env deadline)
)

;;;*******************************
;;; Contract generation
;;;*******************************
;;; Parametri:
;;; IN:
;;;   sell: lista de asocciatie ordonata crescator cu ofertele de vanzare
;;;	buy:  lista de asocciatie ordonata crescator cu ofertele de cumparare
;;;   M+1_price: pretul de tranzactionare calculat de functia clear_auction    
;;;   good: produsul pentru care s-a organizat licitatia
;;;   env: mediul care a organizat-o
;;;   deadline: in cat timp contractul trebuie executat
;;; OUT:
;;;  lista de contracte de forma
;;; ((a1.sell_price) (a2.buy_price) good_name price t_issue, t_maturity))
;;; Algoritm:
;;; 1) se elimina cumparatorii cu oferte M+1_price
;;; 2) se elimina vanzatorii cu oferte peste M+1_price (=> sell)
;;; 3) se inverseaza lista cu cumparatorii ramasi (=> buy_reverse)
;;; 4) se fac perechi de agenti: primul din lista sell cu primul din lista 
;;;    buy_reverse (=> contracts= ((buyer . bid1) (seller . bid2) (..)...))
;;; 5) se elimina contractele false care apar; exemplu: 3 vanzatori vand la 
;;;    pretul de 12 si 2 cumparatori la pretul de 14 => ultimul vanzator are un 
;;;    contract fals (desi ofera sa e castigatoare) (=> contracts)
;;; 6) pentru fiecare contract din lista de contracte 
;;;        a) se adauga pe rand: numele produsului, pretul, t_issue, t_maturity
;;;        b) se inverseaza pentru (=>((a1.3)(b.5) g1 p t1 t2)) 
;;;        c) se adauga in lista de contracte active ale agentilor semnatari
;;;        d) se adauga in lista de noi contracte generate in runda curenta (=> 
;;;           new_contracts
;;; 7) se adauga new_contracts la contractele produsului good

(defun contract_generation (sell buy M+1_price good env deadline)
"generate contracts of the form :
((a1.sell_price) (a2.buy_price) good_name price t_issue, t_maturity))
where t_maturity = t_issue + deadline"
    
;elimin cumparatorii sub pretul stabilit
   (for each element in buy do        
	(cond ((< (cdr element) M+1_price) 
			(setf buy (remove element buy)))))
   (format t "~% Potentialii cumparatori sunt: ~S " buy)

;elimin vanzatorii peste pretul stabilit
   (for each element in sell do 
       (cond ((> (cdr element) M+1_price) 
				(setf sell (remove element sell)))))
   (format t "~% Potentialii vanzatori sunt: ~S " sell)

  (setq buy_reverse (reverse buy)) ;pe prima pozitie va fi cel mai mare bid
  ;doar o parte din potentialii cumparatori sau vanzatori vor semna contracte
  
  (setq contracts (mapcar #'list buy_reverse sell)) 
  ;dupa ce lista contracts se va inversa, contractile vor avea forma din lucrare
  ; de aceea initial cumparatorul e inaintea vanzatorului

  (format t "~% Lista contractelor semnate in runda ~A (inclusive cele false): ~S" (environment-step env) contracts)
    ;contine si false contracte formate din (agent nil)
    ; se elimina acestea (doar primele nr_min sunt reale)
  (setq nr_contracts (min (length sell) (length buy)))
  (setf contracts (first_n_elements contracts nr_contracts))
  (format t "~% Lista contractelor reale semnate in runda ~A: ~S "
			(environment-step env)  contracts)
    ;adaug pretul la contract
  (setq new_contracts nil)
  (for each contract in contracts do
        (setf contract (cons (good-name good) contract))
        (setf contract (cons M+1_price contract))
        (setf contract (cons (environment-step env) contract))
        (setf contract (cons (+ (environment-step env) deadline) contract))
        (setf contract (reverse contract))
        (format t "~% Contract semnat: ~S" contract)
        (setf new_contracts (append (list contract) new_contracts))
        
        (setq seller_name (car (first contract)))
        (setq buyer_name (car (second contract)))
        (setq seller_agent (return_agent seller_name (environment-agents env)))
        (setq buyer_agent (return_agent buyer_name (environment-agents env)))
        (add_contract_to_agent seller_agent contract)
        (add_contract_to_agent buyer_agent contract)

        
       (format t "~%Contractele detinute de agentul ~A pentru bunul ~A sunt: ~S"
			(agent-name seller_agent) (good-name good) 
                  (agent-active_contracts seller_agent))
       (format t "~%Contractele detinute de agentul ~A pentru bunul ~A sunt: ~S"
			(agent-name buyer_agent) (good-name good) 
                  (agent-active_contracts buyer_agent))

   )

  (setf (good-contract good) (append (good-contract good) new_contracts))
  ; append deoarece contractile raman active pana la expirare sau breach
  ; (nu se sterg de la o runda la alta) 
  (format t "~% Contractele pentru bunul ~A la tactul ~A sunt: ~S "
                      (good-name good)
                      (environment-step env)
                      (good-contract good))  

)

(defun introduce_Po (good env deadline) 
  (setq g1 good) 
  (setf g1 (MAKE-good :name 'copy
                      :sell_agents (good-sell_agents good)
		          :buy_agents (good-buy_agents good)
                      :sell_bids (good-sell_bids good)
			    :buy_bids (good-buy_bids good)
                      :contract nil))
(format t "~%Bunul initial este: ~S" good)

 (dolist (item (good-sell_bids g1)) ;item (agent. oferta)
     (setf (good-sell_bids g1) (remove item (good-sell_bids g1)))
     (clear_auction g1 env deadline)
     (format t "~%Bunul initial este: ~S" good)
     (format t "~%Bunul g1 este:~S" g1)
     (setf Po_s 0)
     (dolist (contract (good-contract g1))
       (cond ((equal (car item) (get_contract_seller_name contract)) 
                (setf Po_s (get_contract_price contract))))) 
     (dolist (contract (good-contract good))
        (cond ((equal (car item) (get_contract_seller_name contract))
                (setf contract (append contract (list Po_s))))))  


     (setf (good-contract g1) nil))
       
(dolist (item (good-buy_bids g1))
     (setf (good-buy_bids g1) (remove item (good-buy_bids g1)))
     (clear_auction g1 env deadline)
     (format t "~%Bunul initial este: ~S" good)
     (format t "~%Bunul g1 este:~S" g1)
     (setf Po_b 0)
     (dolist (contract (good-contract g1))
       (cond ((equal (car item) (get_contract_buyer_name contract)) 
                (setf Po_b (get_contract_price contract)))))    
     (dolist (contract (good-contract good))
        (cond ((equal (car item) (get_contract_buyer_name contract))
                (setf contract (append contract (list Po_b))))))  
   
     (setf (good-contract g1) nil))




)

