;;; File: env_update.lisp -*- Mode: Lisp; Syntax: Common-Lisp; -*-
;;; Contine functiile:
;;;	(update env)
;;;   (performance agent env)
;;;	(execute-action agent action env)
;;;   (clear_auction good env deadline)
;;;   (contract_generation sell buy M+1_price good env deadline)
;;;	(perform_contracts good env deadline)
;;;   (add-score agent sum)
;;;	(decrease_score agent sum)
;;;   (add_good agent good_name)
;;;   (decrease_good agent good_name)
;;;   (add_contract_to_agent agent contract)
;;;   (remove_contract_from_agent agent contract)


;******************************
; Updating the environment
;******************************
;;; functia update apelata la fiecare pas face urmatoarele lucruri:
;;; 1) sterge ofertele din runda precedenta din variabilele good-sell_bids
;;;   si good-buy_bids pentru fiecare good
;;; 2)pentru fiecare agent colecteaza fiecare actiune a sa prin apelul functiei 
;;;   (execute-agent agent env)
;;; 3)pentru fiecare good se contractele care au tmaturity=step se executa
;;; 4)pentru fiecare good se calculeaza noile contracte generate in runda 
;;;	curenta prin apelul functiei (clear_auction good env deadline)

(defun update (env)
  
	(setq deadline 4)
      (ascii-display-update-fn T env)
	(print "Executing actions now...")
	 
      ;stergerea ofertelor din runda precedenta
      (for each good in (environment-goods env) do
        (setf (good-sell_bids good) nil)
        (setf (good-buy_bids good) nil))
    
      (for each agent in (environment-agents env) do
        (cond ((and (is_supplier agent) (equal (agent-stored_items agent) nil)) 
                      (decrease_score agent (agent-valuation_in agent))
			    (add_good agent (car (agent-output_goods agent))))
               ((and (is_consumer agent) 
                     (not (equal (agent-stored_items agent) nil))) 
                         (add_score agent (agent-valuation_out agent))
			       (decrease_good agent (car (agent-input_goods agent))))

        )
        (for each action in (car (agent-action agent)) do 
       	   (execute-action agent action env)))

      (for each good in (environment-goods env) do
            (perform_contracts good env deadline)
     		(clear_auction good env deadline)
            (introduce_Po good env deadline) 
      )
      
        (setf *total_score* 0) ;scorul tuturor agentilor din mediu
        (format t "~% Scorul la finalul rundei ~A este:" (environment-step env))   	
        (for each agent in (environment-agents env) do
          (format t "~% Agentul ~A are suma ~A si produsele ~S" 
		(agent-name agent) (agent-score agent) (agent-stored_items agent))
          (setf *total_score* (+ *total_score* (agent-score agent))))
        (format t "~%Total welfare is: ~A " *total_score*)
        
        (if (equal T *view_network*) (tk::mkPlot))
       
        (format t "~%View network? ~S" *view_network*)
         	
        (print "_____________________________________")
)


(defun performance (agent env)
  (agent-score agent)
)


(defun execute-action (agent action env)
"Execute the action by applying the function to its arguments,
 if the action is something that the agent can do."
  ;(format t "Action is ~S " action)
  (let ((legal (agent-legal-actions agent)))
    (when (or (eq legal t) (member (op action) legal))
      (apply (op action) agent env (args action))))
  ;(format T "Argumente ~S" (args action))
)

;;;***************************
;;; Performing contracts
;;;***************************

;;; pentru fiecare contract al unui produs care expira in runda curenta 
;;;      a)se extrag agentii din acel contract
;;;      b)daca detine produsul, agentul vanzator primeste banii si cedeaza 
;;;        produsul
;;;      c)agentul cumparator primeste produsul si cedeaza banii
;;;      d)se sterge contractul din lista de contracte ale produsului
;;;      e)se sterge contractul din listele cu contracte active ale agentilor

(defun perform_contracts (good env deadline)
 (dolist  (expired_contract (good-contract good))
    (cond ((equal (environment-step env) (nth 5 expired_contract))
            ;daca tmaturity=tact
              (setq seller_name (car (first expired_contract)))
              (setq buyer_name (car (second expired_contract)))
              (setq seller (return_agent seller_name 
                                    (environment-agents env)))
              (setq buyer (return_agent buyer_name 
                                   (environment-agents env)))
              (cond ((member (good-name good) (agent-stored_items seller))
                          (tranzactie seller buyer expired_contract good)
                    )
                    (T (format t "~% Agentul vanzator ~A nu detine produsul pentru care a semnat" (agent-name seller))
                       (format t "~% Ca urmare se considera ca agentul ~A a incalcat contractul current ~S"
                                  (agent-name seller) expired_contract)
                        ; (breach seller env good (nth 4 expired_contract))
                     (breach_contract seller buyer expired_contract good env)
                    ) 
              )           
           )
     )
 )

 (format t "~%Contractele executate runda aceasta pentru bunul ~A sunt: ~S" 
                 (good-name good) (good-contract good)) 
)


;;;Efectueaza tranzactia intre cei 2 agenti pentru un contract
;;; 	a) agentul vanzator primeste banii
;;;   b) agentului cumparator i se scad banii
;;;   c) agentului vanzator I se scade bunul din lista de bunuri stocate
;;;   d) agentului cumparator I se adauga bunul in lista de bunuri stocate
;;;   e) se sterge contractul din lista de contracte pentru bunul good 
;;;   f) se sterge contractul din lista de contracte active ale agentilor

(defun tranzactie (seller buyer expired_contract good)
    (format t "~%Tranzactie intre ~A si ~A:"
		    (agent-name seller) (agent-name buyer))
    (format t "~%Agentul ~A primeste suma ~A"
                (agent-name seller) (nth 3 expired_contract))
    (add_score seller (nth 3 expired_contract))
    (decrease_score buyer (nth 3 expired_contract))

    (format t "~%Agentul ~A primeste bunul ~A"
                (agent-name buyer) (nth 2 expired_contract))
    (format t "~% Agentul ~S detine bunurile ~S" 
                (agent-name seller) (agent-stored_items seller))
    (decrease_good seller (nth 2 expired_contract))
    (format t "~% Agentul ~S detine bunurile ~S" 
                (agent-name seller) (agent-stored_items seller))

    (add_good buyer (nth 2 expired_contract))
    (setf (good-contract good) 
                (remove expired_contract (good-contract good)))
    (format t "~%Contractele active pentru bunul ~A sunt: ~S"
		    (good-name good) (good-contract good))

    (remove_contract_from_agent seller expired_contract) 
    (remove_contract_from_agent buyer expired_contract) 
  )

(defun add_score (agent sum)
"Creste scorul agentului agent cu valoare sum"
(setf (agent-score agent) (+ (agent-score agent) sum)))

(defun decrease_score (agent sum)
"Micsoreaza scorul agentului agent cu valoare sum"
(setf (agent-score agent) (- (agent-score agent) sum)))

(defun add_good (agent good_name)
"Concateneaza numele produsului in lista de produse stocate de agentul agent"
(setf (agent-stored_items agent) (cons good_name (agent-stored_items agent))))

(defun decrease_good (agent good_name)
"Elimina numele produslui din lista de produse stocate de agentul agent" 
(setf (agent-stored_items agent) (remove good_name (agent-stored_items agent))))


