;;; File: utils.lisp -*- Mode: Lisp; Syntax: Common-Lisp; -*-
;;; Functii continute:
;;;        -definirea structurii good
;;;        -(enumerate_goods env)
;;;	     -(add_input_goods agent input_goods env)
;;;        -(add_output_goods agent output_goods env)
;;;        -(return_good nume lista_goods)
;;;        -(return_agent nume lista_agents)
;;;        -(first_n_elements ls n)
;;;        -(sort_assoc lista)
      

;*********************
; Aux
;*********************
; mediul este alcatuit practic din aceasta lista de produse
; pentru fiecare produs se initiaza o licitatie

(defstruct (good) 
 (name "G")
 (sell_agents nil)  ;agentii care au declarat la initializare ca vand produsul 
 (buy_agents nil)   ;agentii care au declarat achiztioneaza produsul 
 (sell_bids nil)    ;lista de asociatii cu agentii si ofertele din runda curenta
 (buy_bids nil)     ; -||-
 (contract nil))    ;lista de contracte active
 

;;;;**********************************
;;; Enumerate goods
;;;;**********************************

;;; este prima functie apelata la initializarea mediului
;;; mediul colecteaza de la fiecare agent bunurile pe care acesta le ofera spre 
;;; vanzare sau doreste sa le cumpere
;;; informatiile colectate sunt stocate in lista d ebunuri a mediului

(defun enumerate_goods (env)
  ;(print (environment-agents env))
     (for each agent in (environment-agents env) do
	(add_input_goods agent (agent-input_goods agent) env)
	(add_output_goods agent (agent-output_goods agent) env))
)

;;;Adauga produsele pe care un agent doreste sa le cumpere in mediu si trece 
;;;agentul in lista de doritori ai produsului respectiv 
;;;Algoritm:
;;;   1) creez lista names care contine numele de produse din mediu
;;;   2) daca produsul nu se afla in lista cu numele de produse atunci 
;;;                    a) se creaza un nou obiect de tip good;
;;;                    b) se introduce agentul la lista posibililor cumparatori
;;;      daca produsul exista deja in mediu
;;;                    a) se intoarce bunul cu numele respective
;;;                       (cu functia (return_obeject nume lista)
;;;                    b) se adauga agentul in ista de posibili cumparatori ai 
;;;				  acestui bun

(defun add_input_goods (agent input_goods env)

   (setq names nil)    ;lista cu nemele produselor existente in mediu
   (for each g in (environment-goods env) do
      (setf names (cons (good-name g) names)))

   (for-each good_name in input_goods do     
	(cond ((not (member good_name names))
			(setf (environment-goods env)
				(cons (MAKE-good :name good_name
                 					:sell_agents nil
		     					:buy_agents (list agent)
                 					:contract nil)
                        (environment-goods env)) ))
		;se modifica lista buy_agents pentru bunul cu acelasi nume 
              (T (setf (good-buy_agents 
                       (return_good good_name (environment-goods env))) 
                       (cons agent (good-buy_agents 
                       (return_good good_name (environment-goods env))))))
       )
    )
   ;(print goods)
)


;;;Adauga produsele pe care un agent doreste sa le vanda in mediu si trece 
;;;agentul in lista de ofertanti ai produsului respectiv 
;;;Algoritm:
;;;   1) creez lista names care contine numele de produse din mediu
;;;   2) daca produsul nu se afla in lista cu numele de produse atunci 
;;;                    a) se creaza un nou obiect de tip good;
;;;                    b) se introduce agentul ca prim posibil vanzator
;;;      daca produsul exista deja in mediu
;;;                    a) se intoarce bunul cu numele respectiv
;;;                       (cu functia (return_obeject nume lista)
;;;                    b) se adauga agentul in lista de posibili cumparatori ai 
;;;				  acestui bun


(defun add_output_goods (agent output_goods env)
  (setq names nil)
  (for each g in (environment-goods env) do
   (setf names (cons (good-name g) names)))

  (for-each good_name in output_goods do     
	(cond ((not (member good_name names))
			(setf (environment-goods env)
					 (cons (MAKE-good :name good_name
                 					:sell_agents (list agent)
		     					:buy_agents nil
                 					:contract nil)
                         	 (environment-goods env))
                  )
             )
            (T (setf (good-sell_agents 
			   (return_good good_name (environment-goods env))) 
			   (cons agent (good-sell_agents 
   			   (return_good good_name (environment-goods env))))))
       )
  )
 ;(print goods)
)

;;; Intoarce numele unui obiect
;;; Folosita d efunctia return_good(val list)
(defun return_name (good)
(good-name good)
)

;;; var numele dupa care se cauta in lista de obiecte list list contine 
;;; structuri cu slotul :name
;;; var este numele dupa care se face cautare
;;; intoarce structura de tip good care are numele var"

 (defun return_good (var list)
	(find var list :test #'equal :key #'return_name))


(defun return_agent_name (agent)
(agent-name agent)
)

(defun return_agent (var list)
	(find var list :test #'equal :key #'return_agent_name))


(defun first_n_elements (ls n)
"Intoarce lista formata din primele n elemente din lista ls"
(cond ((or (null ls) (zerop n)) nil)
      (T (cons (first ls) (first_n_elements (rest ls) (- n 1)))))
)

(defun sort_assoc (assoc_list)
"Sorteaza o lista de asociatie dupa campul cdr (valoare)
 -are effect destructive asupra listei originale"
  (sort assoc_list #'< :key #'cdr)
)



